from transformers import FlaxViTForImageClassification
from transformers.models.vit.configuration_vit import ViTConfig
from flax import nnx
from typing import Union
from vitax.models.vit import VisionTransformer
from vitax.models.vit import default_kernel_init

VisionTransformers = ['google/vit-base-patch16-224',
                    'google/vit-large-patch16-224',
                    'google/vit-large-patch16-384',
                    'google/vit-base-patch16-384',
                    'google/vit-base-patch32-384',
                    'google/vit-large-patch32-384',
                    'google/vit-large-patch16-224-in21k',
                    'google/vit-base-patch16-224-in21k',
                    'google/vit-huge-patch14-224-in21k',
                    'google/vit-base-patch32-224-in21k',
                    'google/vit-large-patch32-224-in21k']

def vit_inplace_copy_weights(*, src_model, dst_model, config):

    assert isinstance(src_model, FlaxViTForImageClassification)
    assert isinstance(dst_model, VisionTransformer)

    tf_model_params = src_model.params
    tf_model_params_fstate = nnx.traversals.flatten_mapping(tf_model_params)

    flax_model_params = nnx.state(dst_model, nnx.Param)
    flax_model_params_fstate = dict(flax_model_params.flat_state())

    num_layers = config.num_hidden_layers
    num_heads = config.num_attention_heads
    head_dim = config.hidden_size // num_heads

    params_name_mapping = {
        ("cls_token",): ("vit", "embeddings", "cls_token"),
        ("position_embeddings",): ("vit", "embeddings", "position_embeddings"),
        **{
            ("patch_embeddings", x): ("vit", "embeddings", "patch_embeddings", "projection", x)
            for x in ["kernel", "bias"]
        },
        **{
            ("encoder", "layers", i, "attn", y, x): (
                "vit", "encoder", "layer", str(i), "attention", "attention", y, x
            )
            for x in ["kernel", "bias"]
            for y in ["key", "value", "query"]
            for i in range(num_layers)
        },
        **{
            ("encoder", "layers", i, "attn", "out", x): (
                "vit", "encoder", "layer", str(i), "attention", "output", "dense", x
            )
            for x in ["kernel", "bias"]
            for i in range(num_layers)
        },
        **{
            ("encoder", "layers", i, "mlp", "layers", y1, x): (
                "vit", "encoder", "layer", str(i), y2, "dense", x
            )
            for x in ["kernel", "bias"]
            for y1, y2 in [(0, "intermediate"), (3, "output")]
            for i in range(num_layers)
        },
        **{
            ("encoder", "layers", i, y1, x): (
                "vit", "encoder", "layer", str(i), y2, x
            )
            for x in ["scale", "bias"]
            for y1, y2 in [("norm1", "layernorm_before"), ("norm2", "layernorm_after")]
            for i in range(num_layers)
        },
        **{
            ("final_norm", x): ("vit", "layernorm", x)
            for x in ["scale", "bias"]
        },
        **{
            ("classifier", x): ("classifier", x)
            for x in ["kernel", "bias"]
        }
    }

    nonvisited = set(flax_model_params_fstate.keys())

    for key1, key2 in params_name_mapping.items():
        assert key1 in flax_model_params_fstate, key1
        assert key2 in tf_model_params_fstate, (key1, key2)

        nonvisited.remove(key1)

        src_value = tf_model_params_fstate[key2]
        if key2[-1] == "kernel" and key2[-2] in ("key", "value", "query"):
            shape = src_value.shape
            src_value = src_value.reshape((shape[0], num_heads, head_dim))

        if key2[-1] == "bias" and key2[-2] in ("key", "value", "query"):
            src_value = src_value.reshape((num_heads, head_dim))

        if key2[-4:] == ("attention", "output", "dense", "kernel"):
            shape = src_value.shape
            src_value = src_value.reshape((num_heads, head_dim, shape[-1]))

        dst_value = flax_model_params_fstate[key1]
        assert src_value.shape == dst_value.value.shape, (key2, src_value.shape, key1, dst_value.value.shape)
        dst_value.value = src_value.copy()
        assert dst_value.value.mean() == src_value.mean(), (dst_value.value, src_value.mean())

    assert len(nonvisited) == 0, nonvisited

    nnx.update(dst_model, nnx.State.from_flat_path(flax_model_params_fstate))


def vit_get_model(
    name_or_config: Union[str, dict],
    num_classes: int = 1000,
    pretrained: bool = True,
    fsdp = True
) -> VisionTransformer:

    if pretrained and isinstance(name_or_config, dict):
        raise ValueError("Can not create pretrained model with custom config, please set pretrained=False")

    if isinstance(name_or_config, str) and name_or_config not in VisionTransformers:
        raise ValueError(f"Model name '{name_or_config}' is not a valid or available model.")

    if pretrained:

        hf_model = FlaxViTForImageClassification.from_pretrained(name_or_config)
        config = hf_model.config

        model = VisionTransformer(
            num_classes=config.num_labels,
            img_size=config.image_size,
            patch_size=config.patch_size,
            num_layers=config.num_hidden_layers,
            num_heads=config.num_attention_heads,
            mlp_dim=config.intermediate_size,
            hidden_size=config.hidden_size,
            dropout_rate=config.hidden_dropout_prob,
            rngs=nnx.Rngs(0)
        )

        # Use the inplace function to copy the weights
        vit_inplace_copy_weights(src_model=hf_model, dst_model=model, config=config)

        # If the user specified a different number of classes for a fine-tuning task,
        # replace the classifier head *after* loading all other pre-trained weights.
        if num_classes != config.num_labels:
            model.classifier = nnx.Linear(
                in_features=model.classifier.in_features,
                out_features=num_classes,
                kernel_init= nnx.with_partitioning(
                nnx.initializers.lecun_normal(),
                ('data', None),  # Shard along the input feature axis
            ) if fsdp else default_kernel_init,

                rngs=nnx.Rngs(0)
            )
    else:  # Not pretrained, create a randomly initialized model
        model_kwargs = {}
        if isinstance(name_or_config, str):
            # Get the default configuration for the named model
            config = ViTConfig.from_pretrained(name_or_config)
            model_kwargs = {
                'img_size': config.image_size,
                'patch_size': config.patch_size,
                'num_layers': config.num_hidden_layers,
                'num_heads': config.num_attention_heads,
                'mlp_dim': config.intermediate_size,
                'hidden_size': config.hidden_size,
                'dropout_rate': config.hidden_dropout_prob,
            }
        elif isinstance(name_or_config, dict):
            config_key_mapping = {
                'image_size': 'img_size',
                'patch_size': 'patch_size',
                'num_hidden_layers': 'num_layers',
                'num_attention_heads': 'num_heads',
                'intermediate_size': 'mlp_dim',
                'hidden_size': 'hidden_size',
                'hidden_dropout_prob': 'dropout_rate',
            }
            for key in name_or_config.keys():
                if key in config_key_mapping:
                    model_kwargs[config_key_mapping[key]] = name_or_config[key]
        else:
            raise TypeError(f"Expected 'name_or_config' to be a str or dict, but got {type(name_or_config)}.")

        model = VisionTransformer(
            num_classes=num_classes,
            fsdp=fsdp,
            **model_kwargs,
            rngs=nnx.Rngs(0)
        )

    return model


from typing import Union, Dict, Any, List
import jax
from .vit import VisionTransformer
from .vit_load_weights import vit_get_model , VisionTransformers
from flax import nnx
import optax

@nnx.jit(static_argnames=['mesh', 'optimizer'])
def shard_model_and_create_optimizer(model: VisionTransformer,
                                     optimizer, # Should be an Optax optimizer object
                                     mesh
                                     ):

    optimizer = nnx.optimizer.ModelAndOptimizer(model=model, tx=optimizer)

    model_state = nnx.state(model)
    model_shardings = nnx.get_named_sharding(model_state, mesh)
    model_sharded_state = jax.lax.with_sharding_constraint(model_state, model_shardings)
    nnx.update(model, model_sharded_state)

    optimizer_state = nnx.state(optimizer, nnx.optimizer.OptState)
    optimizer_shardings = nnx.get_named_sharding(optimizer_state, mesh)
    optimizer_sharded_state = jax.lax.with_sharding_constraint(optimizer_state, optimizer_shardings)
    nnx.update(optimizer, optimizer_sharded_state)

    return model, optimizer

class ViTFactory:

    def can_build(self,
                  name_or_config: Union[str, Dict]) -> bool:

        if isinstance(name_or_config, dict):
            return True

        if isinstance(name_or_config, str) and name_or_config in VisionTransformers:
            return True
        return False


    def build(self,
              name_or_config: Union[str, Dict],
              fsdp: bool,
              mesh=None,
              optimizer=None,
              **kwargs: Any):

        if fsdp:

            model = vit_get_model(name_or_config=name_or_config, fsdp=fsdp, **kwargs)

            # This returns sharded model and optimizer, should be unpacked when calling the create_model function
            return shard_model_and_create_optimizer(model=model,
                                                    mesh=mesh,
                                                    optimizer=optimizer)
        else:

            return vit_get_model(name_or_config=name_or_config, fsdp=fsdp, **kwargs)


_MODEL_FACTORIES: List[Any] = [
    ViTFactory(),
]


# --- The Main Public API Function ---

def create_model(
    name_or_config: Union[str, Dict],
    fsdp: bool = False,
    optimizer=None,
    **kwargs: Any,
) -> VisionTransformer:

    if fsdp and not optimizer:
        raise ValueError("For FSDP, you should provide an optax optimizer.")

    if fsdp:
        num_devices = jax.device_count()
        mesh = jax.make_mesh(
            axis_shapes=(num_devices,),
            axis_names=('data',),
        )
    else: 
        mesh = None

    for factory in _MODEL_FACTORIES:

        if factory.can_build(name_or_config):

            return factory.build(name_or_config=name_or_config,
                                 fsdp=fsdp,
                                 mesh=mesh,
                                 optimizer=optimizer,
                                 **kwargs)

    raise ValueError(f"Could not find a model factory for '{name_or_config}'.")


import jax
import jax.numpy as jnp
from flax import nnx
from flax.nnx.nn import dtypes, initializers

default_kernel_init = initializers.lecun_normal()


class VisionTransformer(nnx.Module):
    """ Implements the ViT model, inheriting from `flax.nnx.Module`. """

    def __init__(
            self,
            num_classes: int = 1000,
            in_channels: int = 3,
            img_size: int = 224,
            patch_size: int = 16,
            num_layers: int = 12,
            num_heads: int = 12,
            mlp_dim: int = 3072,
            hidden_size: int = 768,
            dropout_rate: float = 0.1,
            fsdp: bool = True,
            *,
            rngs: nnx.Rngs = nnx.Rngs(0),
    ):
        n_patches = (img_size // patch_size) ** 2

        # We shard the kernel of the convolution layer along the output feature dimension ('hidden_size').
        # This distributes the initial projection across devices.
        self.patch_embeddings = nnx.Conv(
            in_channels,
            hidden_size,
            kernel_size=(patch_size, patch_size),
            strides=(patch_size, patch_size),
            padding="VALID",
            use_bias=True,
            kernel_init=nnx.with_partitioning(
                nnx.initializers.lecun_normal(),
                (None, None, None, 'data'),  # Shard along the output feature axis
            ) if fsdp else default_kernel_init,
            rngs=rngs,
        )

        initializer = jax.nn.initializers.truncated_normal(stddev=0.02)
        self.position_embeddings = nnx.Param(
            initializer(rngs.params(), (1, n_patches + 1, hidden_size), jnp.float32)
        )
        self.dropout = nnx.Dropout(dropout_rate, rngs=rngs)
        self.cls_token = nnx.Param(jnp.zeros((1, 1, hidden_size)))

        self.encoder = nnx.Sequential(*[
            TransformerEncoder(hidden_size, mlp_dim, num_heads,
                               dropout_rate, fsdp=fsdp, rngs=rngs)
            for i in range(num_layers)
        ])
        self.final_norm = nnx.LayerNorm(hidden_size, rngs=rngs)

        # We shard the final classification layer's kernel along its input dimension.
        self.classifier = nnx.Linear(
            hidden_size,
            num_classes,
            kernel_init=nnx.with_partitioning(
                nnx.initializers.lecun_normal(),
                ('data', None),  # Shard along the input feature axis
            ) if fsdp else default_kernel_init,
            rngs=rngs,
        )

    def __call__(self, x: jax.Array) -> jax.Array:
        patches = self.patch_embeddings(x)
        batch_size = patches.shape[0]
        patches = patches.reshape(batch_size, -1, patches.shape[-1])

        cls_token = jnp.tile(self.cls_token, [batch_size, 1, 1])
        x = jnp.concat([cls_token, patches], axis=1)
        embeddings = x + self.position_embeddings
        embeddings = self.dropout(embeddings)

        x = self.encoder(embeddings)
        x = self.final_norm(x)
        x = x[:, 0]

        return self.classifier(x)


class TransformerEncoder(nnx.Module):
    """ A single transformer encoder block in the ViT model. """

    def __init__(
            self,
            hidden_size: int,
            mlp_dim: int,
            num_heads: int,
            dropout_rate: float = 0.0,
            fsdp:bool = True,
            *,
            rngs: nnx.Rngs = nnx.Rngs(0),
    ) -> None:
        self.norm1 = nnx.LayerNorm(hidden_size, rngs=rngs)

        self.attn = nnx.MultiHeadAttention(
            num_heads=num_heads,
            in_features=hidden_size,
            dropout_rate=dropout_rate,
            broadcast_dropout=False,
            decode=False,
            deterministic=False,
            kernel_init=nnx.with_partitioning(
                nnx.initializers.xavier_uniform(),
                ('data', None, None) ,  # Shard input features
            ) if fsdp else default_kernel_init,

            out_kernel_init=nnx.with_partitioning(
                nnx.initializers.xavier_uniform(),
                ('data', None) if fsdp else default_kernel_init,  # Shard output projection
            ) if fsdp else default_kernel_init,
            rngs=rngs,
        )

        self.norm2 = nnx.LayerNorm(hidden_size, rngs=rngs)

        self.mlp = nnx.Sequential(
            nnx.Linear(
                hidden_size,
                mlp_dim,
                kernel_init=nnx.with_partitioning(
                    nnx.initializers.lecun_normal(), (None, 'data')
                ) if fsdp else default_kernel_init,
                rngs=rngs
            ),
            nnx.gelu,
            nnx.Dropout(dropout_rate, rngs=rngs),
            nnx.Linear(
                mlp_dim,
                hidden_size,  # Corrected output dimension
                kernel_init=nnx.with_partitioning(
                    nnx.initializers.lecun_normal(), ('data', None)
                ) if fsdp else default_kernel_init,
                rngs=rngs
            ),
            nnx.Dropout(dropout_rate, rngs=rngs),
        )

    def __call__(self, x: jax.Array) -> jax.Array:
        x = x + self.attn(self.norm1(x))
        x = x + self.mlp(self.norm2(x))
        return x


from flax import nnx
import optax
import jax, jax.numpy as jnp


def compute_losses_and_logits(model: nnx.Module,
                              images: jax.Array,
                              labels: jax.Array):

    logits = model(images)
    loss = optax.softmax_cross_entropy_with_integer_labels(
        logits=logits, labels=labels
    ).mean()
    return loss, logits


@nnx.jit
def train_step(
        model: nnx.Module, optimizer: nnx.Optimizer, batch
):
    images = batch[0]
    labels = batch[1]

    grad_fn = nnx.value_and_grad(compute_losses_and_logits, has_aux=True)
    (loss, logits), grads = grad_fn(model, images, labels)

    optimizer.update(grads)  # In-place updates.

    return loss


@nnx.jit
def eval_step(
        model: nnx.Module, batch, eval_metrics: nnx.MultiMetric
):
    images = batch[0]
    labels = batch[1]
    loss, logits = compute_losses_and_logits(model, images, labels)

    eval_metrics.update(
        loss=loss,
        logits=logits,
        labels=labels,
    )


